from typing import Optional

import torch

from .BaseHSMM import BaseHSMM # type: ignore
from ..emissions import CategoricalEmissions # type: ignore
from ..utils import validate_sequence # type: ignore

class CategoricalHSMM(BaseHSMM, CategoricalEmissions):
    """
    Categorical Hidden semi-Markov Model (HSMM)
    ----------
    Hidden semi-Markov model with categorical (discrete) emissions. This model is an extension of classical HMMs where the duration of each state is modeled by a geometric distribution.
    Duration in each state is modeled by a Categorical distribution with a fixed maximum duration.

    Parameters:
    ----------
    n_states (int):
        Number of hidden states in the model.
    n_emissions (int): 
        Number of emissions in the model.
    seed (int):
        Random seed for reproducibility.
    params_init (bool):
        Whether to initialize the model parameters prior to fitting.
    init_dist (SAMPLING_DISTRIBUTIONS):
        Distribution to use for initializing the model parameters.
    alpha (float):
        Dirichlet concentration parameter for the prior over initial state probabilities and transition probabilities.
    verbose (bool):
        Whether to print progress logs during fitting.
    """
    def __init__(self,
                 n_states: int,
                 n_emissions: int,
                 max_duration: int,
                 alpha: float = 1.0,
                 params_init: bool = False,
                 seed: Optional[int] = None, 
                 device: Optional[torch.device] = None):
        
        BaseHSMM.__init__(self,n_states,max_duration,params_init,alpha,seed,device)
        
        CategoricalEmissions.__init__(self,n_states,n_emissions,params_init,alpha,seed,device)

    def __str__(self):
        return f'CategoricalHSMM(n_states={self.n_states}, max_duration{self.max_duration}, n_emissions={self.n_features})'

    @property
    def params(self):
        """Returns the parameters of the model."""
        return {
            'pi': self.initial_vector.matrix,
            'A': self.transition_matrix.matrix,
            'D': self.duration_matrix.matrix,
            'B': self.emission_matrix.matrix
        }

    @property    
    def n_fit_params(self):
        """Return the number of trainable model parameters."""
        return {
            'states': self.n_states,
            'transitions': self.n_states**2,
            'durations': self.n_states * self.max_duration,
            'emissions': self.n_states * self.n_features
        }
    
    @property
    def dof(self):
        """Returns the degrees of freedom of the model."""
        return self.n_states ** 2 + self.n_states * self.n_features - 1
    
    def check_sequence(self,sequence):
        return validate_sequence(sequence,True,self.n_features)
        
    def map_emission(self,x):
        return CategoricalEmissions.map_emission(self,x)

    def sample_B_params(self,X,seed=None):
        self._emission_matrix = CategoricalEmissions.sample_emissions_params(self,X,seed)

    def update_B_params(self,X,log_gamma,theta):
        gamma = [torch.exp(gamma) for gamma in log_gamma]
        CategoricalEmissions.update_emission_params(self,X,gamma,theta)




from typing import Optional, Literal
import torch

from .BaseHSMM import BaseHSMM # type: ignore
from ..emissions.gaussian_mix import GaussianMixtureEmissions # type: ignore
from ..emissions import GaussianEmissions # type: ignore
from ..utils import validate_sequence # type: ignore


class GaussianHSMM(BaseHSMM, GaussianEmissions):
    """
    Gaussian Hidden Semi-Markov Model (Gaussian HSMM)
    ----------
    This model assumes that the data follows a multivariate Gaussian distribution. 
    The model parameters (initial state probabilities, transition probabilities, duration probabilities,emission means, and emission covariances) 
    are learned using the Baum-Welch algorithm.

    Parameters:
    ----------
    n_states (int):
        Number of hidden states in the model.
    max_duration (int):
        Maximum duration of the states.
    n_features (int):
        Number of features in the emission data.
    n_components (int):
        Number of components in the Gaussian mixture model.
    params_init (bool):
        Whether to initialize the model parameters prior to fitting.
    alpha (float):
        Dirichlet concentration parameter for the prior over initial state probabilities and transition probabilities.
    covariance_type (COVAR_TYPES):
        Type of covariance parameters to use for the emission distributions.
    min_covar (float):
        Floor value for covariance matrices.
    random_state (Optional[int]):
        Random seed to use for reproducible results.
    verbose (bool):
        Whether to print progress logs during fitting.
    """

    COVAR_TYPES = Literal['spherical', 'tied', 'diag', 'full']

    def __init__(self,
                 n_states: int,
                 n_features: int,
                 max_duration: int,
                 params_init: bool = False,
                 k_means: bool = False,
                 alpha: float = 1.0,
                 min_covar: float = 1e-3,
                 covariance_type: COVAR_TYPES = 'full',
                 random_state: Optional[int] = None,
                 device: Optional[torch.device] = None):

        BaseHSMM.__init__(self,n_states,max_duration,params_init,alpha,random_state,device)
        
        GaussianEmissions.__init__(self,n_states,n_features,params_init,k_means,covariance_type,min_covar,random_state,device)
                    
    def __str__(self):
        return f'GaussianHSMM(n_states={self.n_states}, n_durations={self.max_duration},n_features={self.n_features})'

    @property
    def n_fit_params(self):
        """Return the number of trainable model parameters."""
        return {
            'states': self.n_states,
            'transitions': self.n_states**2,
            'durations': self.n_states * self.max_duration,
            'means': self.n_states * self.n_features,
            'covars': {
                'spherical': self.n_states,
                'diag': self.n_states * self.n_features,
                'full': self.n_states * self.n_features * (self.n_features + 1) // 2,
                'tied': self.n_features * (self.n_features + 1) // 2,
            }[self.covariance_type],
        }
    
    @property
    def params(self):
        return {'pi':self.initial_vector.matrix,
                'A':self.transition_matrix.matrix,
                'D':self.duration_matrix.matrix,
                'means': self.means, 
                'covars': self.covs}

    @property
    def dof(self):
        return self.n_states**2 - 1 + self.means.numel() + self.covs.numel() 
    
    def check_sequence(self,sequence):
        return validate_sequence(sequence,False,self.n_features)
    
    def map_emission(self,x):
        return GaussianEmissions.map_emission(self,x)

    def sample_B_params(self,X=None,seed=None):
        self._means,self._covs = GaussianEmissions.sample_emissions_params(self,X,seed)

    def update_B_params(self,X,log_gamma,theta=None):
        gamma = [gamma.exp() for gamma in log_gamma]
        GaussianEmissions.update_emission_params(self,X,gamma,theta)


class GaussianMixtureHSMM(BaseHSMM, GaussianMixtureEmissions):
    """
    Gaussian Hidden Semi-Markov Model (Gaussian HSMM)
    ----------
    This model assumes that the data follow a multivariate Gaussian distribution. 
    The model parameters (initial state probabilities, transition probabilities, duration probabilities,emission means, and emission covariances) 
    are learned using the Baum-Welch algorithm.

    Parameters:
    ----------
    n_states (int):
        Number of hidden states in the model.
    max_duration (int):
        Maximum duration of the states.
    n_features (int):
        Number of features in the emission data.
    n_components (int):
        Number of components in the Gaussian mixture model.
    params_init (bool):
        Whether to initialize the model parameters prior to fitting.
    alpha (float):
        Dirichlet concentration parameter for the prior over initial state probabilities and transition probabilities.
    covariance_type (COVAR_TYPES):
        Type of covariance parameters to use for the emission distributions.
    min_covar (float):
        Floor value for covariance matrices.
    random_state (Optional[int]):
        Random seed to use for reproducible results.
    verbose (bool):
        Whether to print progress logs during fitting.
    """

    COVAR_TYPES = Literal['spherical', 'tied', 'diag', 'full']

    def __init__(self,
                 n_states: int,
                 n_features: int,
                 max_duration: int,
                 n_components: int = 1,
                 params_init: bool = False,
                 k_means: bool = False,
                 alpha: float = 1.0,
                 covariance_type: COVAR_TYPES = 'full',
                 min_covar: float = 1e-3,
                 random_state: Optional[int] = None,
                 device: Optional[torch.device] = None):

        BaseHSMM.__init__(self,n_states,max_duration,params_init,alpha,random_state,device)
        
        GaussianMixtureEmissions.__init__(self,n_states,n_components,n_features,params_init,k_means,alpha,covariance_type,min_covar,random_state,device)
                    
    def __str__(self):
        return f'GaussianMixtureHSMM(n_states={self.n_states}, n_durations={self.max_duration},n_features={self.n_features}, n_components={self.n_components})'

    @property
    def n_fit_params(self):
        """Return the number of trainable model parameters."""
        fit_params_dict = {
            'states': self.n_states,
            'transitions': self.n_states**2,
            'durations': self.n_states * self.max_duration,
            'weights': self.n_states * self.n_components,
            'means': self.n_states * self.n_features * self.n_components,
            'covars': {
                'spherical': self.n_states,
                'diag': self.n_states * self.n_features,
                'full': self.n_states * self.n_features * (self.n_features + 1) // 2,
                'tied': self.n_features * (self.n_features + 1) // 2,
            }[self.covariance_type],
        }

        return fit_params_dict
    
    @property
    def params(self):
        return {'pi':self.initial_vector.matrix,
                'A':self.transition_matrix.matrix,
                'D':self.duration_matrix.matrix,
                'weights': self.weights.matrix,
                'means': self.means, 
                'covars': self.covs}

    @property
    def dof(self):
        return self.n_states**2 - 1 + self.n_states*self.n_components - self.n_states + self.means.numel() + self.covs.numel() 
    
    def check_sequence(self,sequence):
        return validate_sequence(sequence,False,self.n_features)
    
    def map_emission(self,x):
        return GaussianMixtureEmissions.map_emission(self,x)

    def sample_B_params(self,X,seed=None):
        self._means, self._covs = GaussianMixtureEmissions.sample_emissions_params(self,X,seed)

    def update_B_params(self,X,log_gamma,theta=None):
        posterior_vec = []
        resp_vec = GaussianMixtureEmissions._compute_responsibilities(self,X)
        for resp,gamma_val in zip(resp_vec,log_gamma):
            posterior_vec.append(torch.exp(resp + gamma_val.T.unsqueeze(1)))

        GaussianMixtureEmissions.update_emission_params(self,X,posterior_vec,theta)


import torch
from torch.distributions import Categorical # type: ignore
from typing import Optional, List

from .base_emiss import BaseEmission # type: ignore
from ..stochastic_matrix import EmissionMatrix # type: ignore
from ..utils import ContextualVariables, log_normalize # type: ignore


class CategoricalEmissions(BaseEmission):
    """
    Categorical emission distribution for HMMs.

    Parameters:
    ----------
    n_dims (int):
        Number of hidden states in the model.
    n_features (int):
        Number of emissions in the model.
    init_params (bool):
        Whether to initialize the emission parameters.
    alpha (float):
        Dirichlet concentration parameter for the prior over emission probabilities.
    seed (int):
        Random seed for reproducibility.
    device (torch.device):
        Device on which to fit the model.

    Attributes:
    ----------
    emission_matrix (EmissionMatrix):
        Emission matrix representing the categorical distribution.
    """

    def __init__(self,
                 n_dims:int,
                 n_features:int,
                 init_params:bool=True,
                 alpha:float = 1.0,
                 seed:Optional[int] = None,
                 device:Optional[torch.device] = None):
        
        super().__init__(n_dims,n_features,True,seed,device)

        self.alpha = alpha
        if init_params:
            self._emission_matrix = self.sample_emissions_params(seed=seed)

    def __str__(self):
        return f'CategoricalEmissions(n_dims={self.n_dims}, n_features={self.n_features})'

    @property
    def pdf(self) -> Categorical:
        """Compute the probability density of the emission distribution."""
        return self._emission_matrix._dist

    @property
    def emission_matrix(self) -> EmissionMatrix:
        return self._emission_matrix
    
    @emission_matrix.setter
    def emission_matrix(self, matrix:torch.Tensor):
        assert (self.n_dims,self.n_features) == matrix.shape, f'Expected matrix shape {(self.n_dims,self.n_features)} but got {matrix.shape}'
        if isinstance(matrix, EmissionMatrix):
            self._emission_matrix = matrix
        elif isinstance(matrix, torch.Tensor):
            self._emission_matrix = EmissionMatrix(n_states=self.n_dims, 
                                                   n_emissions=self.n_features, 
                                                   matrix=matrix,
                                                   device=self.device)
        else:
            raise NotImplementedError('Matrix type not supported')

    def sample_emissions_params(self,X=None,seed=None):
        if X is not None:
            emission_freqs = torch.bincount(X) / X.shape[0]
            return EmissionMatrix.from_tensor(emission_freqs.expand(self.n_dims,-1).log())
        else:
            return EmissionMatrix(n_states=self.n_dims,
                                  n_emissions=self.n_features,
                                  rand_seed=seed,
                                  alpha=self.alpha,
                                  device=self.device)

    def map_emission(self,x):
        return self.pdf.log_prob(x.repeat(self.n_dims,1).T)

    def update_emission_params(self,X,gamma,theta=None):
        self._emission_matrix.matrix.copy_(self._compute_emprobs(X,gamma,theta))

    def _compute_emprobs(self, 
                        X:List[torch.Tensor],
                        gamma:List[torch.Tensor],
                        theta:Optional[ContextualVariables]) -> torch.Tensor:  
        """Compute the emission probabilities for each hidden state."""
        emission_mat = torch.zeros(size=(self.n_dims, self.n_features),
                                   dtype=torch.float64,
                                   device=self.device)

        for seq,gamma_val in zip(X,gamma):
            if theta is not None:
                #TODO: Implement contextualized emissions
                raise NotImplementedError('Contextualized emissions not implemented for CategoricalEmissions')
            else:
                masks = seq.view(1,-1) == torch.arange(self.n_features).view(-1,1)
                for i,mask in enumerate(masks):
                    masked_gamma = gamma_val[mask]
                    emission_mat[:,i] += masked_gamma.sum(dim=0)

        return log_normalize(emission_mat.log(),1)


from .categorical import CategoricalEmissions
from .gaussian import GaussianEmissions
from .gamma import GammaEmissions
from .poisson import PoissonEmissions
from .multinomial import MultinomialEmissions
from .gaussian_mix import GaussianMixtureEmissions


import torch
import numpy as np
from typing import Optional, List, Literal, Dict
from .emissions import GaussianMixtureEmissions
from .utils import FittedModel, ConvergenceHandler, INFORM_CRITERIA, validate_sequence


class GaussianMixtureModel(GaussianMixtureEmissions):

    def __init__(self, 
                 n_dims: int,
                 n_components: int,
                 n_features: int,
                 params_init: bool = True,
                 alpha: float = 1.0,
                 covariance_type: GaussianMixtureEmissions.COVAR_TYPES_HINT = 'full',
                 min_covar: float = 1e-3,
                 seed: Optional[int] = None,
                 device: Optional[torch.device] = None):
        
        # TODO: for now K-means are not implemented
        GaussianMixtureEmissions.__init__(self,n_dims,n_components,n_features,params_init,False,alpha,covariance_type,min_covar,seed,device)

    def __str__(self):
        return f'GaussianMixtureModel(n_states={self.n_dims}, n_features={self.n_features}, n_components={self.n_components})'
    
    @property
    def dof(self) -> int:
        return self.n_dims*(self.n_components-1) + self.n_dims*self.n_features*self.n_components + self.n_dims*self.n_features*(self.n_features + 1) // 2
    
    @property
    def n_fit_params(self) -> Dict[str,int]:
        return {'weights': self.n_dims * self.n_components,
                'means': self.n_dims * self.n_features * self.n_components,
                'covs': {
                    'spherical': self.n_dims,
                    'diag': self.n_dims * self.n_features,
                    'full': self.n_dims * self.n_features * (self.n_features + 1) // 2,
                    'tied': self.n_features * (self.n_features + 1) // 2,
                }[self.covariance_type]}
    
    def score(self,X:torch.Tensor) -> float:
        return self.map_emission(X).sum(0).item()

    def ic(self, 
           X:torch.Tensor, 
           criterion:Literal['AIC','BIC','HQC'] = 'AIC') -> float:
        """Calculates the information criteria for a given model."""
        log_likelihood = self.score(X)
        if criterion not in INFORM_CRITERIA:
            raise NotImplementedError(f'{criterion} is not a valid information criterion. Valid criteria are: {INFORM_CRITERIA}')
        
        criterion_compute = {'AIC': lambda log_likelihood, dof: -2.0 * log_likelihood + 2.0 * dof,
                             'BIC': lambda log_likelihood, dof: -2.0 * log_likelihood + dof * np.log(X.shape[0]),
                             'HQC': lambda log_likelihood, dof: -2.0 * log_likelihood + 2.0 * dof * np.log(np.log(X.shape[0]))}[criterion]
        
        return criterion_compute(log_likelihood, self.dof)

    def fit(self,
            X:torch.Tensor,
            tol:float=1e-2,
            max_iter:int=20,
            n_init:int=1,
            post_conv_iter:int=3,
            ignore_conv:bool=False,
            plot_conv:bool=False,
            verbose:bool=False,
            lengths:Optional[List[int]] = None,
            theta:Optional[torch.Tensor] = None) -> Dict[int,FittedModel]:
        """Fit the model to the data."""
        X_valid = validate_sequence(X,False,self.n_features)
        X_vec = list(torch.split(X_valid, lengths)) if lengths is not None else [X_valid]

        self.conv = ConvergenceHandler(tol=tol,
                                       max_iter=max_iter,
                                       n_init=n_init,
                                       post_conv_iter=post_conv_iter,
                                       device=self.device,
                                       verbose=verbose)

        distinct_models = {}
        for rank in range(n_init):
            self.conv.push_pull(self.score(X),0,rank)
            for iter in range(1,self.conv.max_iter+1):
                # EM algorithm step
                self._update_params(X=X_vec,
                                    resp=self._compute_responsibilities(X_vec),
                                    theta=theta)
                
                curr_log_like = self.score(X)
                converged = self.conv.push_pull(curr_log_like,iter,rank)
                if converged and not ignore_conv:
                    print(f'Model converged after {iter} iterations with log-likelihood: {curr_log_like:.2f}')
                    break

            distinct_models[rank] = FittedModel(self.__str__(),
                                                self.n_fit_params, 
                                                self.dof,
                                                converged,
                                                curr_log_like, 
                                                self.ic(X),
                                                self.params)
        
        if plot_conv:
            self.conv.plot_convergence()

        return distinct_models


#TODO: this module might be redudant, consider removing it
#TODO: this init method is too confusing, supply matrix, states, and emissions or their combination and inheritance together 
from abc import ABC, abstractmethod, abstractproperty
from typing import Optional, Union, Sequence, TypeVar, Tuple

import torch
import numpy as np

from .utils import SeedGenerator, print_table, states_names, sample_prob_matrix, validate_prob_matrix # type: ignore

torch.set_printoptions(precision=4, profile="full")
MAT_OPS = TypeVar('MAT_OPS', bound=Union['StochasticMatrix', np.ndarray, torch.Tensor])

class StochasticMatrix(ABC):
    """ 
    Stochastic Matrix
    ----------
    This class is used to represent a stochastic matrix. It is a wrapper around a torch.Tensor with a
    torch.distributions.Categorical distribution.

    Parameters:
    ----------
    matrix (Categorical):
        The stochastic matrix.
    seed (Optional[int]):
        Random seed to use for reproducible results.
    
    Attributes:
    ----------
    matrix (torch.Tensor):
        The stochastic matrix.
    seed (Optional[int]):
        Random seed to use for reproducible results.
    """
    
    def __init__(self,
                 matrix_size: Tuple[int,...],
                 matrix: Optional[torch.Tensor] = None,
                 alpha: float = 1.0,
                 semi_markov: bool = False,
                 seed: Optional[int] = None,
                 device: Optional[torch.device] = None):

        self.seed_generator = SeedGenerator(seed)
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') if device is None else device
        
        mat = sample_prob_matrix(alpha, self.device, matrix_size,semi_markov) if matrix is None else matrix
        self._dist = torch.distributions.Categorical(logits=mat)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(shape={self.matrix.shape})"

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractproperty
    def view(self):
        pass

    @property
    def seed(self):
        return self.seed_generator.seed

    @property
    def matrix(self) -> torch.Tensor:
        return self._dist.logits

    @matrix.setter
    def matrix(self, matrix: Union[torch.Tensor, np.ndarray]):
        assert self.matrix.shape == matrix.shape, "Matrix must be the same shape as the original"
        if isinstance(matrix, np.ndarray):
            self._dist.logits = torch.from_numpy(matrix)
        elif isinstance(matrix, torch.Tensor):
            self._dist.logits = matrix
        else:
            raise NotImplementedError(f'Expected one of Tensor or numpy array, got {type(matrix)}')
        
    def __mul__(self, other: MAT_OPS) -> torch.Tensor:
        if isinstance(other, StochasticMatrix):
            return self.matrix * other.matrix
        elif isinstance(other, torch.Tensor):
            return self.matrix * other
        elif isinstance(other, np.ndarray):
            return self.matrix * torch.from_numpy(other)
        else:
            raise NotImplementedError(f'Unexpected type for operation, got {type(other)}')
        
    def __rmul__(self, other: MAT_OPS) -> torch.Tensor:
        return self.__mul__(other)
    
    def __matmul__(self, other: MAT_OPS) -> torch.Tensor:
        if isinstance(other, StochasticMatrix):
            return self.matrix @ other.matrix
        elif isinstance(other, torch.Tensor):
            return self.matrix @ other
        elif isinstance(other, np.ndarray):
            return self.matrix @ torch.from_numpy(other)
        else:
            raise NotImplementedError(f'Unexpected type for operation, got {type(other)}')
    
    def __rmatmul__(self, other: MAT_OPS) -> torch.Tensor:
        return self.__matmul__(other)

    def __truediv__(self, other: MAT_OPS) -> torch.Tensor:
        if isinstance(other, StochasticMatrix):
            return self.matrix / other.matrix
        elif isinstance(other, torch.Tensor):
            return self.matrix / other
        elif isinstance(other, np.ndarray):
            return self.matrix / torch.from_numpy(other)
        else:
            raise NotImplementedError(f'Unexpected type for operation, got {type(other)}')
    
    def __rtruediv__(self, other: MAT_OPS) -> torch.Tensor:
        return self.__truediv__(other)

    def __add__(self, other: MAT_OPS) -> torch.Tensor:
        if isinstance(other, StochasticMatrix):
            return self.matrix + other.matrix
        elif isinstance(other, torch.Tensor):
            return self.matrix + other
        elif isinstance(other, np.ndarray):
            return self.matrix + torch.from_numpy(other)
        else:
            raise NotImplementedError(f'Unexpected type for operation, got {type(other)}')
        
    def __radd__(self, other: MAT_OPS) -> torch.Tensor:
        return self.__add__(other)
    
    def __sub__(self, other: MAT_OPS) -> torch.Tensor:
        if isinstance(other, StochasticMatrix):
            return self.matrix - other.matrix
        elif isinstance(other, torch.Tensor):
            return self.matrix - other
        elif isinstance(other, np.ndarray):
            return self.matrix - torch.from_numpy(other)
        else:
            raise NotImplementedError(f'Unexpected type for operation, got {type(other)}')
        
    def __rsub__(self, other: MAT_OPS) -> torch.Tensor:
        return self.__sub__(other)

    def __eq__(self, other: MAT_OPS) -> bool: #type: ignore
        if isinstance(other, StochasticMatrix):
            return torch.equal(self.matrix, other.matrix)
        elif isinstance(other, torch.Tensor):
            return torch.equal(self.matrix, other)
        elif isinstance(other, np.ndarray):
            return torch.equal(self.matrix, torch.from_numpy(other))
        else:
            raise NotImplementedError(f'Unexpected type for operation, got {type(other)}')
    
    def __len__(self) -> int:
        return self.matrix.numel()
    
    def __getitem__(self, idx) -> torch.Tensor:
        if isinstance(idx, tuple):
            row_idx, col_idx = idx
            if isinstance(row_idx, slice) or isinstance(col_idx, slice):
                return self.matrix[row_idx, col_idx]
            else:
                return self.matrix[row_idx, col_idx]
        else:
            return self.matrix[idx]
    
    def __setitem__(self, row_idx, col_idx, value):
        self.matrix[row_idx, col_idx] = value


class EmissionMatrix(StochasticMatrix):
    """
    Emission Matrix
    ---------------
    A Class representing the Emission matrix of a finite Markov chain.
    """

    def __init__(self,
                 n_states: int,
                 n_emissions: int,
                 matrix: Optional[torch.Tensor]=None,
                 states: Optional[Sequence[str]]=None,
                 emissions: Optional[Sequence[str]]= None,
                 rand_seed: Optional[int] = None,
                 alpha: float = 1.0,
                 device: Optional[torch.device] = None):
        
        self.n_states = n_states
        self.n_emissions = n_emissions
        self.states = states_names(n=n_states, state_type='state') if states is None else states
        self.emissions = states_names(n=n_emissions, state_type='emission') if emissions is None else emissions

        super().__init__(matrix=matrix,
                         matrix_size=(n_states, n_emissions),
                         alpha=alpha,
                         seed=rand_seed,
                         device=device)
        
    def __str__(self) -> str:
        return f"EmissionMatrix(n_states = {self.n_states}, n_emissions = {self.n_emissions})"

    @property
    def states(self) -> Sequence[str]:
        return self._states

    @states.setter
    def states(self, states: Sequence[str]):
        unique_names = len(set(states))

        if unique_names != self.n_states:
            raise ValueError(f'Expected {self.n_states} unique states, got {unique_names}')
        else:
            self._states = states

    @property
    def emissions(self) -> Sequence[str]:
        return self._emissions

    @emissions.setter
    def emissions(self, emissions: Sequence[str]):
        unique_names = len(set(emissions))
        
        if unique_names != self.n_emissions:
            raise ValueError(f'Expected {self.n_emissions} unique emissions, got {unique_names}')
        else:
            self._emissions = emissions

    @classmethod
    def from_tensor(cls, matrix:torch.Tensor, log:bool=True):
        (n_states, n_emissions) = matrix.shape
        valid_prob_matrix = validate_prob_matrix(matrix,log)
        log_prob_matrix = torch.log(valid_prob_matrix) if not log else valid_prob_matrix 
        return cls(n_states,n_emissions, log_prob_matrix)

    def view(self, log_scale: bool = False):
        vals = self.matrix.exp() if not log_scale else self.matrix
        vals_list = vals.round().tolist()
        
        trans = [[self.states[idx]] + row for idx, row in enumerate(vals_list)]

        print_table(title='Emission Matrix',
                    header=['State\Emission'] + list(self.emissions), 
                    rows=trans)

class TransitionMatrix(StochasticMatrix):
    """
    Transition Matrix
    --------
    A Class representing the Transition matrix of a finite Markov chain.

    Parameters
    ----------
    n_states : int
        Number of states in the system.
    matrix : Optional[torch.Tensor], optional
        An initial probability distribution matrix, by default `None`.
    states : Optional[List[str]], optional
        A list of names/labels for each state in the system, by default `None`.
    rand_seed : Optional[int], optional
        The seed for the random number generator, by default `None`.
    init_dist : Optional[SAMPLING_DISTRIBUTIONS], optional
        The distribution to use when initializing the matrix if not provided, by default `'Uniform'`.
    alpha : Optional[float], optional
        The prior to use when initializing the matrix if not provided, by default `1.0`.

    Note
    ----
    If `matrix` is provided, it will be used as the initial probability distribution. 
    If `matrix` is not provided, a distribution specified by `init_dist` will be used to initialize the matrix.
    """

    def __init__(self,
                 n_states: int,
                 matrix: Optional[torch.Tensor] = None,
                 states: Optional[Sequence[str]]= None,
                 rand_seed: Optional[int] = None,
                 alpha: float = 1.0,
                 semi_markov: bool = False,
                 device: Optional[torch.device] = None):
        
        self.n_states = n_states
        self.states = states_names(n=n_states, state_type='state') if states is None else states
        self.semi_markov = semi_markov

        super().__init__(matrix=matrix,
                         matrix_size=(n_states, n_states),
                         semi_markov=semi_markov,
                         alpha=alpha,
                         seed=rand_seed,
                         device=device)

    def __str__(self) -> str:
        return f"TransitionMatrix(n_states = {self.n_states})"

    @property
    def states(self) -> Sequence[str]:
        return self._states

    @states.setter
    def states(self, states: list):
        if len(set(states)) != self.n_states:
            raise ValueError(f'Expected {self.n_states} states, got {len(set(states))}')
        else:
            self._states = states

    @classmethod
    def from_tensor(cls, matrix:torch.Tensor, log:bool=True):
        (n_states, n_states_2) = matrix.shape
        valid_prob_matrix = validate_prob_matrix(matrix,log)
        log_prob_matrix = torch.log(valid_prob_matrix) if not log else valid_prob_matrix 
        assert n_states == n_states_2, f'Expected square matrix, got {log_prob_matrix.shape}'
        return cls(n_states, log_prob_matrix)

    def view(self, log_scale: bool = False):
        vals = self.matrix.exp() if not log_scale else self.matrix
        vals_list = vals.round().tolist()
        
        trans = [[self.states[idx]] + row for idx, row in enumerate(vals_list)]

        print_table(title='Transition Matrix',
                    header=['State\State'] + list(self.states), 
                    rows=trans)


class DurationMatrix(StochasticMatrix):
    """
    Duration Matrix
    ---------------
    A Class representing the Duration matrix of a semi-Markov chain.
    """

    def __init__(self,
                 n_states: int,
                 max_dur: int,
                 matrix: Optional[torch.Tensor] = None,
                 states: Optional[Sequence[str]]= None,
                 durations: Optional[Sequence[str]]= None,
                 rand_seed: Optional[int] = None,
                 alpha: float = 1.0,
                 device: Optional[torch.device] = None):
        
        self.n_states = n_states
        self.max_duration = max_dur
        self.states = states_names(n=n_states, state_type='state') if states is None else states
        self.durations = states_names(n=max_dur, state_type='duration') if durations is None else durations

        super().__init__(matrix=matrix,
                         matrix_size=(n_states, max_dur),
                         alpha=alpha,
                         seed=rand_seed,
                         device=device)
        
    def __str__(self) -> str:
        return f"DurationMatrix(states = {self.n_states}, max_duration = {self.max_duration})"

    @property
    def durations(self) -> Sequence[str]:
        return self._durations

    @durations.setter
    def durations(self, durations: list):
        if len(set(durations)) != self.max_duration:
            raise ValueError(f'Expected {self.max_duration} unique durations, got {len(set(durations))}')
        else:
            self._durations = durations

    @property
    def states(self) -> Sequence[str]:
        return self._states

    @states.setter
    def states(self, states: list):
        if len(set(states)) != self.n_states:
            raise ValueError(f'Expected {self.n_states} unique states, got {len(set(states))}')
        else:
            self._states = states

    @classmethod
    def from_tensor(cls, matrix:torch.Tensor, log:bool=True):
        (n_states, max_dur) = matrix.shape
        valid_prob_matrix = validate_prob_matrix(matrix,log)
        log_prob_matrix = torch.log(valid_prob_matrix) if not log else valid_prob_matrix 
        return cls(n_states, max_dur,log_prob_matrix)

    def view(self, log_scale: bool = False):
        vals = self.matrix.exp() if not log_scale else self.matrix
        vals_list = vals.round().tolist()
        
        trans = [[self.states[idx]] + row for idx, row in enumerate(vals_list)]

        print_table(title='Duration Matrix',
                    header=['State\Emission'] + list(self.durations), 
                    rows=trans)


class WeightsMatrix(StochasticMatrix):
    """
    Weights Matrix
    ---------------
    A Class representing the Weights matrix of a Gaussian Mixture HMM.
    """

    def __init__(self,
                 n_states: int,
                 n_components: int,
                 matrix: Optional[torch.Tensor] = None,
                 states: Optional[Sequence[str]] = None,
                 components: Optional[Sequence[str]] = None,
                 rand_seed: Optional[int] = None,
                 alpha: float = 1.0,
                 device: Optional[torch.device] = None):
        
        self.n_states = n_states
        self.n_components = n_components
        self.states = states_names(n=n_states, state_type='state') if states is None else states
        self.components = states_names(n=n_components, state_type='component') if components is None else components

        super().__init__(matrix=matrix,
                         matrix_size=(n_states, n_components),
                         alpha=alpha,
                         seed=rand_seed,
                         device=device)
        
    def __str__(self) -> str:
        return f"WeightsMatrix(n_states = {self.n_states}, n_components = {self.n_components})"

    @property
    def states(self) -> Sequence[str]:
        return self._states

    @states.setter
    def states(self, states: list):
        if len(set(states)) != self.n_states:
            raise ValueError(f'Expected {self.n_states} unique states, got {len(set(states))}')
        else:
            self._states = states

    @property
    def components(self) -> Sequence[str]:
        return self._components
    
    @components.setter
    def components(self, components: list):
        if len(set(components)) != self.n_components:
            raise ValueError(f'Expected {self.n_components} unique components, got {len(set(components))}')
        else:
            self._components = components

    def view(self, log_scale: bool = False):
        vals = self.matrix.exp() if not log_scale else self.matrix
        vals_list = vals.round().tolist()
        
        trans = [[self.states[idx]] + row for idx, row in enumerate(vals_list)]

        print_table(title='Weights Matrix',
                    header=['State\Component'] + list(self.components), 
                    rows=trans)
        
    @classmethod
    def from_tensor(cls, matrix:torch.Tensor, log:bool=True):
        (n_states, n_components) = matrix.shape
        valid_prob_matrix = validate_prob_matrix(matrix,log)
        log_prob_matrix = torch.log(valid_prob_matrix) if not log else valid_prob_matrix 
        return cls(n_states, n_components,log_prob_matrix)


class ProbabilityVector(StochasticMatrix):
    """
    Probability Vector
    ------------------
    A class used to represent a probability vector.

    Parameters
    ----------
    n_states : int
        Number of states in the system.
    vector : Optional[torch.Tensor], optional
        An initial probability distribution vector, by default `None`.
    states : Optional[List[str]], optional
        A list of names/labels for each state in the system, by default `None`.
    rand_seed : Optional[int], optional
        The seed for the random number generator, by default `None`.
    init_dist : Optional[SAMPLING_DISTRIBUTIONS], optional
        The distribution to use when initializing the vector if not provided, by default `'Uniform'`.
    alpha : Optional[float], optional
        The prior to use when initializing the vector if not provided, by default `1.0`.

    Note
    ----
    If `vector` is provided, it will be used as the initial probability distribution. 
    If `vector` is not provided, a distribution specified by `init_dist` will be used to initialize the vector.
    """

    def __init__(self,
                 n_states: int,
                 vector: Optional[torch.Tensor] = None,
                 states: Optional[Sequence[str]]= None,
                 rand_seed: Optional[int] = None,
                 alpha: float = 1.0,
                 device: Optional[torch.device] = None):
        
        self.n_states = n_states
        self.states = states_names(n=n_states, state_type='state') if states is None else states

        super().__init__(matrix=vector,
                         matrix_size=(n_states,),
                         alpha=alpha,
                         seed=rand_seed,
                         device=device)

    def __str__(self) -> str:
        return f"ProbabilityVector(n_states = {self.n_states})"

    @property
    def states(self) -> Sequence[str]:
        return self._states

    @states.setter
    def states(self, states: list):
        if len(set(states)) == self.n_states:
            self._states = states
        else:
            raise ValueError(f'Expected {self.n_states} unique states, got {len(set(states))}')
        
    @classmethod
    def from_tensor(cls, matrix:torch.Tensor, log:bool=True):
        (n_states,) = matrix.shape
        valid_prob_matrix = validate_prob_matrix(matrix,log)
        log_prob_matrix = torch.log(valid_prob_matrix) if not log else valid_prob_matrix 
        return cls(n_states, log_prob_matrix)
    
    def view(self, log_scale: bool = False):
        row_name = 'Log Prob' if log_scale else 'Prob'
        vals = self.matrix.exp() if not log_scale else self.matrix
        vals_list = vals.round().tolist()
        
        print_table(title='Initial Distribution',
                    header=['State'] + list(self.states), 
                    rows=[[row_name] + vals_list])
        


from typing import Generator, Sequence, Tuple, Optional, List, Generator, Dict
from dataclasses import dataclass, field

import torch
import matplotlib.pyplot as plt # type: ignore
from prettytable import PrettyTable # type: ignore

DECODERS = frozenset(('viterbi', 'map'))
INFORM_CRITERIA = frozenset(('AIC', 'BIC', 'HQC'))


@dataclass
class FittedModel:
    """Dataclass for a fitted model"""
    model: str
    n_fit_params: Dict[str,int]
    DOF: int
    converged: bool
    log_likelihood: float
    AIC: float
    params: Dict[str,torch.Tensor]

@dataclass
class Observations:
    """Dataclass for a sequence of observations."""
    n_samples: int
    X: List[torch.Tensor]
    log_probs: List[torch.Tensor]
    lengths: List[int]
    n_sequences: int = field(default=1)

@dataclass
class ContextualVariables:
    """Dataclass for contextual variables."""
    n_context: int
    matrix: List[torch.Tensor]
    time_dependent: bool = field(default=False)

class SeedGenerator:
    def __init__(self, seed: Optional[int] = None):
        if seed is not None:
            self.seed_gen = torch.random.manual_seed(seed)
        else:
            initial_seed = torch.random.seed()
            self.seed_gen = torch.random.manual_seed(initial_seed)

    def __repr__(self) -> str:
        return f"SeedGenerator(seed={self()})"

    def __call__(self) -> int:
        return self.seed_gen.seed()
    
    @property
    def seed(self) -> int:
        return self.seed_gen.initial_seed()
    
    @seed.setter
    def seed(self, seed: int) -> None:
        self.seed_gen = torch.random.manual_seed(seed)


class ConvergenceHandler:
    """
    Convergence Monitor
    ----------
    Convergence monitor for HMM training. Stores the score at each iteration and checks for convergence.

    Parameters
    ----------
    max_iter : int
        Maximum number of iterations.
    n_init : int
        Number of initializations.
    tol : float
        Convergence threshold.
    post_conv_iter : int
        Number of iterations to run after convergence.
    verbose : bool
        Print convergence information.
    """

    def __init__(self, 
                 max_iter:int,
                 n_init:int, 
                 tol:float, 
                 post_conv_iter:int,
                 device:torch.device, 
                 verbose:bool = True):
        
        self.tol = tol
        self.verbose = verbose
        self.post_conv_iter = post_conv_iter
        self.device = device
        self.max_iter = max_iter
        self.score = torch.full(size=(max_iter+1,n_init),
                                fill_value=float('nan'), 
                                dtype=torch.float64,
                                device=self.device)
        self.delta = self.score.clone()

    def __repr__(self):
        return f"""
        ConvergenceHandler(tol={self.tol},
                            n_iters = {self.iter+1},
                            post_conv_iter={self.post_conv_iter},
                            converged={self.converged},
                            verbose={self.verbose})
                """
    
    def push_pull(self, new_score:float, iter:int, rank:int) -> bool:
        """Push a new score and check for convergence."""
        self.push(new_score, iter, rank)
        return self.converged(iter, rank)
        
    def push(self, new_score:float, iter:int, rank:int):
        """Update the iteration count."""
        self.score[iter,rank] = new_score
        self.delta[iter,rank] = new_score - self.score[iter-1,rank]

    def converged(self, iter:int, rank:int) -> bool:
        """Check if the model has converged and update the convergence monitor."""
        conv_lag = iter-self.post_conv_iter

        if conv_lag < 0:
            self.is_converged = False
        elif all(self.delta[conv_lag:iter, rank] < self.tol):
            self.is_converged = True
        else:
            self.is_converged = False

        if self.verbose:
            score = self.score[iter,rank].item()
            delta = self.delta[iter,rank].item()

            if iter == 0:
                print(f"Run {rank+1} | Initialization | Score: {score:.2f}")
            else:
                print(f"Run {rank+1} | " +
                    f"Iteration: {iter} | " + 
                    f"Score: {score:.2f} | " +
                    f"Delta: {delta:.2f} | " +
                    f"Converged = {self.is_converged}"
                    )

        return self.is_converged
    
    def plot_convergence(self):
        # Define input for plot
        labels = [f'Log-likelihood - Run #{i+1}' for i in range(self.score.shape[1])]

        # Plot setting
        plt.style.use('ggplot')
        _, ax = plt.subplots(figsize=(10, 7))
        ax.plot(torch.arange(self.max_iter+1), 
                self.score.cpu(), 
                linewidth=2, 
                marker='o', 
                markersize=5, 
                label=labels)
        
        ax.set_title('HMM Model Log-Likelihood Convergence')
        ax.set_xlabel('# Iterations')
        ax.set_ylabel('Log-likelihood')
        ax.legend(loc='lower right')
        plt.show()

def states_names(n: int, state_type: str) -> Sequence[str]:
    type_alias = state_type[0].upper()
    return [f'{type_alias}_{n}' for n in range(n)]

def laplace_smoothing(log_matrix: torch.Tensor, k: float) -> torch.Tensor:
    """Laplace smoothing of a log probability matrix"""

    if k <= 0.0:  
        return log_matrix
    else:
        real_matrix = torch.exp(log_matrix) 
        smoothed_log_matrix = torch.log((real_matrix + k) / (1 + k * real_matrix.shape[-1]))
        return smoothed_log_matrix

def validate_prob_matrix(matrix:torch.Tensor, log:bool=True) -> torch.Tensor:    
    if matrix.ndim > 2:
        raise ValueError(f"Matrix must be at most 2-dimensional, got {matrix.ndim} instead")
    elif torch.any(torch.isnan(matrix)):
        raise ValueError("Matrix must not contain NaNs")
    elif log and torch.any(torch.isinf(matrix)):
        raise ValueError("Real matrix must not contain infinities")
    elif not log and torch.any(1 <= matrix <= 0):
        raise ValueError("Real matrix must be between 0 and 1")
    elif log and torch.any(0 <= matrix):
        raise ValueError("Log matrix must be negative")
    else:
        return matrix

def sample_prob_matrix(prior: float, device:torch.device, target_size: Tuple[int,...], semi: bool = False) -> torch.Tensor:
    """Initialize a matrix of probabilities"""
    alphas = torch.full(size=target_size,
                        fill_value=prior,
                        dtype=torch.float64,
                        device=device)
    
    probs = torch.distributions.Dirichlet(alphas).sample()
    if semi:
        probs.fill_diagonal_(0)
        probs /= probs.sum(dim=-1, keepdim=True)

    return probs.log()

def log_normalize(matrix: torch.Tensor, dim:int=0) -> torch.Tensor:
    """Normalize a posterior probability matrix"""
    return matrix - matrix.logsumexp(dim,True)

def sequence_generator(X: Observations) -> Generator[Tuple[int,torch.Tensor,torch.Tensor], None, None]:
    for X_len,seq,log_probs in zip(X.lengths,X.X,X.log_probs):
        yield X_len, seq, log_probs

def validate_sequence(sequence:torch.Tensor, 
                      discrete:bool, 
                      n_features:int) -> torch.Tensor:
    """Do basic checks on sequence dimensions and values. Return a dataclass with the sequence and its properties."""
    if discrete:
        if (n_dim:=sequence.ndim) != 1:
            raise ValueError(f'Sequence must be 1-dimensional, got shape {n_dim}')
        elif sequence.dtype != torch.long:
            raise ValueError(f'Sequence must be of type torch.long, got {sequence.dtype}')
        elif torch.max(sequence).item() > (n_features-1):
            raise ValueError('Invalid emission in sequence')
        else:
            return sequence
    else:
        if sequence.ndim != 2:
            raise ValueError(f'Sequence must have shape (T,{n_features}). Got {sequence.shape}.')
        elif sequence.dtype != torch.double:
            raise ValueError(f'Sequence must be of type torch.float, got {sequence.dtype}')
        elif sequence.shape[1] != n_features:
            raise ValueError(f'Sequence must have shape (T,{n_features}). Got {sequence.shape}.')
        else:
            return sequence           

def validate_means(means: torch.Tensor, n_states: int, n_features: int, n_components: Optional[int]=None) -> torch.Tensor:
    """Do basic checks on matrix mean sizes and values"""
    valid_shape = (n_states, n_features) if n_components is None else (n_states, n_components, n_features)

    if (n_dim:=means.ndim) != (v_dim:=len(valid_shape)):
        raise ValueError(f"Tensor must be {v_dim}D, got {n_dim}D instead")
    elif (m_shape:=means.shape) != valid_shape:
        raise ValueError(f"Tensor must have shape {valid_shape}, got {m_shape} instead")
    elif torch.any(torch.isnan(means)):
        raise ValueError("means must not contain NaNs")
    elif torch.any(torch.isinf(means)):
        raise ValueError("means must not contain infinities")
    else:
        return means
    
def validate_lambdas(lambdas: torch.Tensor, n_states: int, n_features: int) -> torch.Tensor:
    """Do basic checks on matrix mean sizes and values"""
    
    if len(lambdas.shape) != 2:
        raise ValueError("lambdas must have shape (n_states, n_features)")
    elif lambdas.shape[0] != n_states:
        raise ValueError("lambdas must have shape (n_states, n_features)")
    elif lambdas.shape[1] != n_features:
        raise ValueError("lambdas must have shape (n_states, n_features)")
    elif torch.any(torch.isnan(lambdas)):
        raise ValueError("lambdas must not contain NaNs")
    elif torch.any(torch.isinf(lambdas)):
        raise ValueError("lambdas must not contain infinities")
    elif torch.any(lambdas <= 0):
        raise ValueError("lambdas must be positive")
    else:
        return lambdas

def validate_covars(covars: torch.Tensor, 
                    covariance_type: str, 
                    n_states: int, 
                    n_features: int,
                    n_components: Optional[int]=None) -> torch.Tensor:
    """Do basic checks on matrix covariance sizes and values"""
    if n_components is None:
        valid_shape = torch.Size((n_states, n_features, n_features))
    else:
        valid_shape = torch.Size((n_states, n_components, n_features, n_features))    

    if covariance_type == 'spherical':
        if len(covars) != n_features:
            raise ValueError("'spherical' covars have length n_features")
        elif torch.any(covars <= 0): 
            raise ValueError("'spherical' covars must be positive")
    elif covariance_type == 'tied':
        if covars.shape[0] != covars.shape[1]:
            raise ValueError("'tied' covars must have shape (n_dim, n_dim)")
        elif (not torch.allclose(covars, covars.T) or torch.any(covars.symeig(eigenvectors=False).eigenvalues <= 0)):
            raise ValueError("'tied' covars must be symmetric, positive-definite")
    elif covariance_type == 'diag':
        if len(covars.shape) != 2:
            raise ValueError("'diag' covars must have shape (n_features, n_dim)")
        elif torch.any(covars <= 0):
            raise ValueError("'diag' covars must be positive")
    elif covariance_type == 'full':
        if len(covars.shape) != 3:
            raise ValueError("'full' covars must have shape (n_features, n_dim, n_dim)")
        elif covars.shape[1] != covars.shape[2]:
            raise ValueError("'full' covars must have shape (n_features, n_dim, n_dim)")
        for n, cv in enumerate(covars):
            eig_vals, _ = torch.linalg.eigh(cv)
            if (not torch.allclose(cv, cv.T) or torch.any(eig_vals <= 0)):
                raise ValueError(f"component {n} of 'full' covars must be symmetric, positive-definite")
    else:
        raise NotImplementedError(f"This covariance type is not implemented: {covariance_type}")
    
    return covars
       
def init_covars(tied_cv: torch.Tensor, 
                covariance_type: str, 
                n_states: int) -> torch.Tensor:
    """Initialize covars to a given covariance type"""

    if covariance_type == 'spherical':
        return tied_cv.mean() * torch.ones((n_states,))
    elif covariance_type == 'tied':
        return tied_cv
    elif covariance_type == 'diag':
        return tied_cv.diag().unsqueeze(0).expand(n_states, -1)
    elif covariance_type == 'full':
        return tied_cv.unsqueeze(0).expand(n_states, -1, -1)
    else:
        raise NotImplementedError(f"This covariance type is not implemented: {covariance_type}")
    
def fill_covars(covars: torch.Tensor, 
                covariance_type: str, 
                n_states: int, 
                n_features: int,
                n_components: Optional[int]=None) -> torch.Tensor:
    """Fill in missing values for covars"""
    
    if covariance_type == 'full':
        return covars
    elif covariance_type == 'diag':
        return torch.stack([torch.diag(covar) for covar in covars])
    elif covariance_type == 'tied':
        return covars.unsqueeze(0).expand(n_states, -1, -1)
    elif covariance_type == 'spherical':
        eye = torch.eye(n_features).unsqueeze(0)
        return eye * covars.unsqueeze(-1).unsqueeze(-1)
    else:
        raise NotImplementedError(f"This covariance type is not implemented: {covariance_type}")

def print_table(rows: list, header: list, title: str):
    """
    Helper method for the pretty print function. It prints the parameters
    as a nice table.
    """
    t = PrettyTable(title=title, 
                    field_names=header, 
                    header_style='upper',
                    padding_width=1, 
                    title_style='upper')
    
    for row in rows:
        t.add_row(row)
    
    print(t)




import torch
from torch.distributions import MultivariateNormal # type: ignore
from typing import Optional, Literal, Tuple, List
from sklearn.cluster import KMeans # type: ignore

from .base_mixture import MixtureEmissions # type: ignore
from ..stochastic_matrix import WeightsMatrix # type: ignore
from ..utils import ContextualVariables, validate_means, validate_covars, fill_covars # type: ignore
    

class GaussianMixtureEmissions(MixtureEmissions):
    """
    Gaussian Mixture Model for HMM emissions.    
    
    Parameters:
    ----------
    n_dims (int):
        Number of mixtures in the model. This is equal to the number of hidden states in the HMM.
    n_components (int):
        Number of components in the mixture model.
    n_features (int):
        Number of features in the data.
    init_weights (bool):
        Whether to initialize the mixture weights prior to fitting.
    init_dist ():
        Distribution to use for initializing the mixture weights.
    alpha (float):
        Dirichlet concentration parameter for the prior over mixture weights.
    k_means (bool):
        Whether to initialize the mixture means using K-Means clustering.
    min_covar (float):
        Minimum covariance for the mixture components.
    covariance_type (COVAR_TYPES):
        Type of covariance matrix to use for the mixture components. One of 'spherical', 'tied', 'diag', 'full'.
    dims (Sequence[str]):
        Names of the mixtures in the model.
    seed (int):
        Random seed for reproducibility.
    device (torch.device):
        Device to use for computations.
    """

    COVAR_TYPES_HINT = Literal['spherical', 'tied', 'diag', 'full']

    def __init__(self, 
                 n_dims: int,
                 n_features: int,
                 n_components: int,
                 params_init: bool = True,
                 k_means: bool = False,
                 alpha: float = 1.0,
                 covariance_type: COVAR_TYPES_HINT = 'full',
                 min_covar: float = 1e-3,
                 seed: Optional[int] = None,
                 device: Optional[torch.device] = None):
        
        MixtureEmissions.__init__(self,n_dims,n_features,n_components,alpha,params_init,seed,device)
        
        self.k_means = k_means
        self.min_covar = min_covar
        self.covariance_type = covariance_type
        if params_init:
            self._weights, self._means, self._covs = self.sample_emissions_params(seed=seed)
            
    @property
    def means(self) -> torch.Tensor:
        return self._means
    
    @means.setter
    def means(self, means: torch.Tensor):
        valid_means = validate_means(means, self.n_dims, self.n_features, self.n_components)
        self._means = valid_means.to(self.device)

    @property
    def covs(self) -> torch.Tensor:
        return self._covs

    @covs.setter
    def covs(self, new_covars: torch.Tensor):
        """Setter function for the covariance matrices."""
        valid_covars = validate_covars(new_covars, self.covariance_type, self.n_dims, self.n_features, self.n_components)
        self._covs = fill_covars(valid_covars, self.covariance_type, self.n_dims, self.n_features, self.n_components).to(self.device)

    @property
    def pdf(self) -> MultivariateNormal:
        return MultivariateNormal(self.means,self.covs)
    
    @property
    def params(self):
        return {'weights': self.weights.matrix,
                'means': self.means, 
                'covs': self.covs}
    
    def sample_emissions_params(self,X=None,seed=None) -> Tuple[WeightsMatrix,torch.Tensor,torch.Tensor]:
        """Initialize the emission parameters."""
        new_weights = self.sample_weights(seed)
        if X is not None:
            means = self._sample_kmeans(X,seed) if self.k_means else X.mean(dim=0,keepdim=True).expand(self.n_dims,self.n_components,-1).clone()
            centered_data = X - X.mean(dim=0)
            covs = (torch.mm(centered_data.T, centered_data) / (X.shape[0] - 1)).expand(self.n_dims,self.n_components,-1,-1).clone()
        else:
            means = torch.zeros(size=(self.n_dims, self.n_components, self.n_features), 
                                dtype=torch.float64, 
                                device=self.device)
        
            covs = self.min_covar + torch.eye(n=self.n_features, 
                                              dtype=torch.float64,
                                              device=self.device).expand((self.n_dims, self.n_components, self.n_features, self.n_features)).clone()

        return new_weights, means, covs

    def update_emission_params(self,X,resp,theta=None):
        self._weights.matrix.copy_(self._compute_weights(resp))
        self._means.copy_(self._compute_means(X,resp,theta)) 
        self._covs.copy_(self._compute_covs(X,resp,theta))
    
    def _sample_kmeans(self,
                       X:torch.Tensor,
                       seed:Optional[int]=None) -> torch.Tensor:
        """Sample cluster means from K Means algorithm"""
        k_means_alg = KMeans(n_clusters=self.n_dims, 
                               random_state=seed, 
                               n_init="auto").fit(X)
        return torch.from_numpy(k_means_alg.cluster_centers_).reshape(self.n_dims,self.n_components,self.n_features)
    
    def _compute_means(self,
                       X:List[torch.Tensor], 
                       responsibilities:List[torch.Tensor],
                       theta:Optional[ContextualVariables]=None) -> torch.Tensor:
        """Compute the means for each hidden state"""
        new_mean = torch.zeros(size=(self.n_dims,self.n_components,self.n_features), 
                               dtype=torch.float64, 
                               device=self.device)
        
        denom = torch.zeros(size=(self.n_dims,self.n_components,1), 
                            dtype=torch.float64, 
                            device=self.device)
        
        for seq,resp in zip(X,responsibilities):
            if theta is not None:
                # TODO: matmul shapes are inconsistent
                raise NotImplementedError('Contextual variables are not yet implemented')
            else:
                new_mean += resp @ seq
                denom += resp.sum(dim=-1,keepdim=True)

        return new_mean / denom
    
    def _compute_covs(self,
                      X:List[torch.Tensor],
                      responsibilities:List[torch.Tensor],
                      theta:Optional[ContextualVariables]=None) -> torch.Tensor:
        """Compute the covariances for each component."""
        new_covs = torch.zeros(size=(self.n_dims,self.n_components,self.n_features,self.n_features), 
                               dtype=torch.float64, 
                               device=self.device)
        
        denom = torch.zeros(size=(self.n_dims,self.n_components,1,1), 
                            dtype=torch.float64, 
                            device=self.device)

        for seq,resp in zip(X,responsibilities):
            if theta is not None:
                # TODO: matmul shapes are inconsistent 
                raise NotImplementedError('Contextualized emissions not implemented for GaussianEmissions')
            else:
                resp_expanded = resp.unsqueeze(-1)
                diff = seq.unsqueeze(0).expand(self.n_dims,self.n_components,-1,-1) - self.means.unsqueeze(2)
                new_covs += torch.transpose(diff * resp_expanded,2,3) @ diff
                denom += torch.sum(resp_expanded,dim=-2,keepdim=True)

        new_covs /= denom
        new_covs += self.min_covar * torch.eye(self.n_features, device=self.device)
        
        return new_covs
    


import torch
from abc import abstractmethod, abstractproperty, ABC
from typing import Optional, List
from ..utils import ContextualVariables


class BaseEmission(ABC):

    def __init__(self,
                 n_dims: int,
                 n_features: int,
                 discrete: bool,
                 seed:Optional[int] = None,
                 device:Optional[torch.device] = None):

        self.n_dims = n_dims
        self.n_features = n_features
        self.discrete = discrete
        self.seed = seed
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') if device is None else device
        
    @abstractproperty
    def __str__(self):
        pass

    @abstractproperty
    def pdf(self):
        pass

    @abstractmethod
    def map_emission(self, x:torch.Tensor) -> torch.Tensor:
        """Convert emissions into log probabilities."""
        pass

    @abstractmethod
    def sample_emissions_params(self, X:Optional[torch.Tensor]=None, seed:Optional[int]=None):
        """Sample emission parameters."""
        pass

    @abstractmethod
    def update_emission_params(self, X:List[torch.Tensor], gamma:List[torch.Tensor], theta:Optional[ContextualVariables]=None):
        """Update emission parameters in the model."""
        pass


from typing import Optional, Sequence, Tuple, Dict, List, Literal
from abc import ABC, abstractmethod, abstractproperty

import torch
import numpy as np

from ..stochastic_matrix import TransitionMatrix, ProbabilityVector # type: ignore
from ..utils import FittedModel,ContextualVariables, ConvergenceHandler, Observations, log_normalize, sequence_generator, DECODERS, INFORM_CRITERIA # type: ignore


class BaseHMM(ABC): 
    """
    Base Abstract Class for HMM
    ----------
    Base Class of Hidden Markov Models (HMM) class that provides a foundation for building specific HMM models.
    """

    def __init__(self,
                 n_states: int,
                 params_init: bool = False,
                 alpha: float = 1.0,
                 seed: Optional[int] = None,
                 device: Optional[torch.device] = None):

        self.n_states = n_states
        self.alpha = alpha
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') if device is None else device
        if params_init:
            self._initial_vector, self._transition_matrix = self.sample_chain_params(self.alpha,seed)

    @property
    def transition_matrix(self) -> TransitionMatrix:
        return self._transition_matrix

    @transition_matrix.setter
    def transition_matrix(self, matrix:torch.Tensor):
        assert (self.n_states,self.n_states) == matrix.shape, 'Matrix dimensions differ from HMM model'
        if isinstance(matrix, TransitionMatrix):
            self._transition_matrix = matrix
        elif isinstance(matrix, torch.Tensor):
            self._transition_matrix = TransitionMatrix(n_states=self.n_states,
                                                       matrix=matrix,
                                                       device=self.device)
        else:
            raise NotImplementedError('Matrix type not supported')

    @property
    def initial_vector(self) -> ProbabilityVector:
        return self._initial_vector

    @initial_vector.setter
    def initial_vector(self, vector:torch.Tensor):
        assert (self.n_states,) == vector.shape, 'Matrix dimensions differ from HMM model'
        if isinstance(vector, ProbabilityVector):
            self._initial_vector = vector
        elif isinstance(vector, torch.Tensor):
            self._initial_vector = ProbabilityVector(n_states=self.n_states, 
                                                     vector=vector,
                                                     device=self.device)
        else:
            raise NotImplementedError('Matrix type not supported')
        
    @property    
    def view_params(self):
        """Print the model parameters."""
        for param in self.params.values():
            param.view()
            print('\n')
        
    @property
    def _check_params(self):
        """Check if the model parameters are set."""
        return self.params
    
    @abstractproperty
    def __str__(self):
        pass

    @abstractproperty
    def params(self) -> Dict[str, torch.Tensor]:
        """Returns the parameters of the model."""
        pass

    @abstractproperty   
    def n_fit_params(self) -> Dict[str, int]:
        """Return the number of trainable model parameters."""
        pass
    
    @abstractproperty
    def dof(self) -> int:
        """Returns the degrees of freedom of the model."""
        pass

    @abstractmethod
    def _update_B_params(self, X:List[torch.Tensor], log_gamma:List[torch.Tensor], theta:Optional[ContextualVariables]):
        """Update the emission parameters."""
        pass

    @abstractmethod
    def check_sequence(self, sequence:torch.Tensor) -> torch.Tensor:
        """Get emission probabilities for a given sequence of observations."""
        pass

    @abstractmethod
    def map_emission(self, x:torch.Tensor) -> torch.Tensor:
        """Get emission probabilities for a given sequence of observations."""
        pass

    @abstractmethod
    def sample_B_params(self, X:Optional[torch.Tensor]=None, seed:Optional[int]=None):
        """Sample the emission parameters."""
        pass
    
    def _forward(self, X:Observations) -> List[torch.Tensor]:
        """Forward pass of the forward-backward algorithm."""
        alpha_vec = []
        for seq_len,_,log_probs in sequence_generator(X):
            log_alpha = torch.zeros(size=(seq_len,self.n_states), 
                                    dtype=torch.float64,
                                    device=self.device)
            
            log_alpha[0] = self._initial_vector + log_probs[0]
            for t in range(1,seq_len):
                log_alpha[t] = torch.logsumexp(log_alpha[t-1].reshape(-1,1) + self._transition_matrix, dim=0) + log_probs[t]

            alpha_vec.append(log_alpha)

        return alpha_vec
    
    def _backward(self, X:Observations) -> List[torch.Tensor]:
        """Backward pass of the forward-backward algorithm."""
        beta_vec = []
        for seq_len,_,log_probs in sequence_generator(X):
            log_beta = torch.zeros(size=(seq_len,self.n_states), 
                               dtype=torch.float64,
                               device=self.device)
            
            for t in reversed(range(seq_len-1)):
                log_beta[t] = torch.logsumexp(self._transition_matrix + log_probs[t+1] + log_beta[t+1], dim=1)
            
            beta_vec.append(log_beta)

        return beta_vec
    
    def _gamma(self, log_alpha:List[torch.Tensor], log_beta:List[torch.Tensor]) -> List[torch.Tensor]:
        """Compute the log-Gamma variable in Hidden Markov Model."""
        gamma_vec = []
        for alpha,beta in zip(log_alpha,log_beta):
            gamma_vec.append(log_normalize(alpha+beta,1))

        return gamma_vec
    
    def _xi(self, X:Observations, log_alpha:List[torch.Tensor], log_beta:List[torch.Tensor]) -> List[torch.Tensor]:
        """Compute the log-Xi variable in Hidden Markov Model."""
        xi_vec = []
        for (seq_len,_,log_probs),alpha,beta in zip(sequence_generator(X),log_alpha,log_beta):
            log_xi = torch.zeros(size=(seq_len-1, self.n_states, self.n_states),
                                 dtype=torch.float64, 
                                 device=self.device)
            
            for t in range(seq_len-1):
                log_xi[t] = alpha[t].reshape(-1,1) + self._transition_matrix + log_probs[t+1] + beta[t+1]

            log_xi -= alpha[-1].logsumexp(dim=0)
            xi_vec.append(log_xi)

        return xi_vec
    
    def _compute_fwd_bwd(self, X:Observations) -> Tuple[List[torch.Tensor],List[torch.Tensor]]:
        log_alpha = self._forward(X)
        log_beta = self._backward(X)
        return log_alpha, log_beta

    def _compute_posteriors(self, X:Observations) -> Tuple[List[torch.Tensor], List[torch.Tensor]]:
        """Execute the forward-backward algorithm and compute the log-Gamma and log-Xi variables."""
        log_alpha, log_beta = self._compute_fwd_bwd(X)
        log_gamma = self._gamma(log_alpha,log_beta)
        log_xi = self._xi(X,log_alpha,log_beta)
        return log_gamma, log_xi
    
    def _accum_pi(self, log_gamma:List[torch.Tensor]) -> torch.Tensor:
        """Accumulate the statistics for the initial vector."""
        log_pi = torch.zeros(size=(self.n_states,),
                             dtype=torch.float64, 
                             device=self.device)

        for gamma in log_gamma:
            log_pi += gamma[0].exp()
        
        return log_normalize(log_pi.log(),0)

    def _accum_A(self, log_xi:List[torch.Tensor]) -> torch.Tensor:
        """Accumulate the statistics for the transition matrix."""
        log_A = torch.zeros(size=(self.n_states,self.n_states),
                             dtype=torch.float64, 
                             device=self.device)
        
        for xi in log_xi:
            log_A += xi.exp().sum(dim=0)

        return log_normalize(log_A.log(),1)
    
    def _update_model(self, X:Observations, theta:Optional[ContextualVariables]) -> float:
        """Compute the updated parameters for the model."""
        log_gamma, log_xi = self._compute_posteriors(X)

        self._initial_vector.matrix.copy_(self._accum_pi(log_gamma))
        self._transition_matrix.matrix.copy_(self._accum_A(log_xi))
        self._update_B_params(X.X,log_gamma,theta)

        return sum(self._compute_log_likelihood(X))
    
    def _viterbi(self, X:Observations) -> Sequence[torch.Tensor]:
        """Viterbi algorithm for decoding the most likely sequence of hidden states."""
        viterbi_path_list = []
        for seq_len,_,log_probs in sequence_generator(X):
            viterbi_path = torch.empty(size=(seq_len,), 
                                       dtype=torch.int32,
                                       device=self.device)
            
            viterbi_prob = torch.empty(size=(self.n_states, seq_len), 
                                       dtype=torch.float64,
                                       device=self.device)
            psi = viterbi_prob.clone()

            # Initialize t=1
            viterbi_prob[:,0] = self._initial_vector + log_probs[0]
            for t in range(1,seq_len):
                trans_seq = viterbi_prob[:,t-1] + log_probs[t]
                trans_seq = self._transition_matrix + trans_seq.reshape((-1, 1))
                viterbi_prob[:,t] = torch.max(trans_seq, dim=0).values
                psi[:,t] = torch.argmax(trans_seq, dim=0)

            # Backtrack the most likely sequence
            viterbi_path[-1] = torch.argmax(viterbi_prob[:,-1])
            for t in reversed(range(seq_len-1)):
                viterbi_path[t] = psi[viterbi_path[t+1],t+1]

            viterbi_path_list.append(viterbi_path)

        return viterbi_path_list
    
    def _map(self, X:Observations) -> List[torch.Tensor]:
        """Compute the most likely (MAP) sequence of indiviual hidden states."""
        map_paths = []
        for gamma,_ in self._compute_posteriors(X):
            map_paths.append(torch.argmax(gamma, dim=1))
        return map_paths

    def _compute_log_likelihood(self, X:Observations) -> List[float]:
        """Compute the log-likelihood of the given sequence."""
        scores = []
        for alpha in self._forward(X):
            scores.append(alpha[-1].logsumexp(dim=0).item())

        return scores
    
    def sample_chain_params(self, alpha:float, seed:Optional[int]=None) -> Tuple[ProbabilityVector, TransitionMatrix]:
        """Initialize the model parameters."""
        return (ProbabilityVector(self.n_states, rand_seed=seed, alpha=alpha, device=self.device), 
                TransitionMatrix(self.n_states, rand_seed=seed, alpha=alpha, device=self.device))

    def to_observations(self, X:torch.Tensor, lengths:Optional[List[int]]=None) -> Observations:
        """Convert a sequence of observations to an Observations object."""
        X_valid = self.check_sequence(X)
        n_samples = X_valid.shape[0]
        seq_lenghts = [n_samples] if lengths is None else lengths
        X_vec = list(torch.split(X_valid, seq_lenghts))

        log_probs = []
        for seq in X_vec:
            log_probs.append(self.map_emission(seq))
        
        return Observations(n_samples,X_vec,log_probs,seq_lenghts,len(seq_lenghts))  
    
    def check_theta(self, theta:torch.Tensor, X:Observations) -> ContextualVariables:
        """Returns the parameters of the model."""
        if (n_dim:=theta.ndim) != 2:
            raise ValueError(f'Context must be 2-dimensional. Got {n_dim}.')
        elif theta.shape[1] not in (1, X.n_samples):
            raise ValueError(f'Context must have shape (context_vars, 1) for time independent context or (context_vars,{X.n_samples}) for time dependent. Got {theta.shape}.')
        else:
            n_context, n_observations = theta.shape
            time_dependent = n_observations == X.n_samples
            adj_theta = torch.vstack((theta, torch.ones(size=(1,n_observations),
                                                        dtype=torch.float64,
                                                        device=self.device)))
            if not time_dependent:
                adj_theta = adj_theta.expand(n_context+1, X.n_samples)

            context_matrix = list(torch.split(adj_theta,X.lengths,1))
            return ContextualVariables(n_context, context_matrix, time_dependent) 

    def fit(self,
            X:torch.Tensor,
            tol:float=1e-2,
            max_iter:int=20,
            n_init:int=1,
            post_conv_iter:int=3,
            ignore_conv:bool=False,
            sample_B_from_X:bool=False,
            verbose:bool=True,
            plot_conv:bool=False,
            lengths:Optional[List[int]]=None,
            theta:Optional[torch.Tensor]=None) -> Dict[int, FittedModel]:
        """Fit the model to the given sequence using the EM algorithm."""
        if sample_B_from_X:
            self.sample_B_params(X)
        X_valid = self.to_observations(X,lengths)
        valid_theta = self.check_theta(theta,X_valid) if theta is not None else None

        self.conv = ConvergenceHandler(tol=tol,
                                       max_iter=max_iter,
                                       n_init=n_init,
                                       post_conv_iter=post_conv_iter,
                                       device=self.device,
                                       verbose=verbose)

        self._check_params
        distinct_models = {}
        for rank in range(n_init):
            if rank > 0:
                self._initial_vector, self._transition_matrix = self.sample_chain_params(self.alpha)
            
            self.conv.push_pull(sum(self._compute_log_likelihood(X_valid)),0,rank)
            for iter in range(1,self.conv.max_iter+1):
                # EM algorithm step
                self._update_model(X_valid, valid_theta)

                # remap emission probabilities after update of B
                X_valid.log_probs = [self.map_emission(x) for x in X_valid.X]
                
                curr_log_like = sum(self._compute_log_likelihood(X_valid))
                converged = self.conv.push_pull(curr_log_like,iter,rank)
                if converged and not ignore_conv:
                    print(f'Model converged after {iter} iterations with log-likelihood: {curr_log_like:.2f}')
                    break

            distinct_models[rank] = FittedModel(self.__str__(),
                                                self.n_fit_params, 
                                                self.dof,
                                                converged,
                                                curr_log_like, 
                                                self.ic(X,lengths),
                                                self.params)
        
        if plot_conv:
            self.conv.plot_convergence()

        return distinct_models

    def predict(self, 
                X:torch.Tensor, 
                lengths:Optional[List[int]] = None,
                algorithm:Literal['map','viterbi'] = 'viterbi') -> Tuple[float, Sequence[torch.Tensor]]:
        """Predict the most likely sequence of hidden states."""
        self._check_params
        if algorithm not in DECODERS:
            raise ValueError(f'Unknown decoder algorithm {algorithm}')
        
        decoder = {'viterbi': self._viterbi,
                   'map': self._map}[algorithm]
        
        X_valid = self.to_observations(X, lengths)
        log_score = sum(self._compute_log_likelihood(X_valid))
        decoded_path = decoder(X_valid)

        return log_score, decoded_path

    def _score_observations(self, X:torch.Tensor, lengths:Optional[List[int]]=None) -> List[float]:
        """Compute the log-likelihood for each sample sequence."""
        self._check_params
        return self._compute_log_likelihood(self.to_observations(X, lengths))

    def score(self, X:torch.Tensor, lengths:Optional[List[int]]=None) -> float:
        """Compute the joint log-likelihood"""
        return sum(self._score_observations(X,lengths))
    
    def score_samples(self, X:torch.Tensor, lengths:Optional[List[int]]=None) -> List[float]:
        """Compute the log-likelihood for each sequence"""
        return self._score_observations(X,lengths)

    def ic(self, 
           X:torch.Tensor, 
           lengths:Optional[List[int]] = None, 
           criterion:Literal['AIC','BIC','HQC'] = 'AIC') -> float:
        """Calculates the information criteria for a given model."""
        log_likelihood = self.score(X, lengths)
        if criterion not in INFORM_CRITERIA:
            raise NotImplementedError(f'{criterion} is not a valid information criterion. Valid criteria are: {INFORM_CRITERIA}')
        
        criterion_compute = {'AIC': lambda log_likelihood, dof: -2.0 * log_likelihood + 2.0 * dof,
                             'BIC': lambda log_likelihood, dof: -2.0 * log_likelihood + dof * np.log(X.shape[0]),
                             'HQC': lambda log_likelihood, dof: -2.0 * log_likelihood + 2.0 * dof * np.log(np.log(X.shape[0]))}[criterion]
        
        return criterion_compute(log_likelihood, self.dof)








"""
ChadHMM
======

Ultra Chad Implementation of Hidden Markov Models in Pytorch (available only to true sigma males)

But seriously this package needs you to help me make it better. I'm not a professional programmer, I'm just a guy who likes to code. 
If you have any suggestions, please let me know. I'm open to all ideas.
"""

# Import HMM objects
from .hmm import CategoricalHMM, GaussianHMM, GaussianMixtureHMM
from .hsmm import CategoricalHSMM, GaussianHSMM, GaussianMixtureHSMM
from .MixtureModels import GaussianMixtureModel
from .stochastic_matrix import TransitionMatrix, EmissionMatrix, TransitionMatrix, WeightsMatrix, DurationMatrix, ProbabilityVector


from typing import Optional
import torch

from .BaseHMM import BaseHMM # type: ignore
from ..emissions import CategoricalEmissions # type: ignore
from ..utils import validate_sequence # type: ignore

class CategoricalHMM(BaseHMM, CategoricalEmissions):
    """
    Categorical Hidden Markov Model (HMM)
    ----------
    Hidden Markov model with categorical (discrete) emissions. This model is a special case of the HSMM model with a geometric duration distribution.

    Parameters:
    ----------
    n_states (int):
        Number of hidden states in the model.
    n_features (int): 
        Number of emissions in the model.
    random_state (int):
        Random seed for reproducibility.
    params_init (bool):
        Whether to initialize the model parameters.
    alpha (float):
        Dirichlet concentration parameter for the prior over initial distribution, transition amd emission probabilities.
    verbose (bool):
        Whether to print progress logs during fitting.
    device (torch.device):
        Device on which to fit the model.
    """

    def __init__(self,
                 n_states: int,
                 n_features: int,
                 alpha: float = 1.0,
                 params_init: bool = False,
                 seed: Optional[int] = None, 
                 device: Optional[torch.device] = None):
        
        BaseHMM.__init__(self,n_states,params_init,alpha,seed,device)
        
        CategoricalEmissions.__init__(self,n_states,n_features,params_init,alpha,seed,device)
        
    def __str__(self):
        return f'CategoricalHMM(n_states={self.n_states}, n_feautures={self.n_features})'

    @property
    def params(self):
        return {
            'pi': self.initial_vector.matrix,
            'A': self.transition_matrix.matrix,
            'B': self.emission_matrix.matrix
        }

    @property
    def n_fit_params(self):
        return {
            'initial_states': self.n_states,
            'transitions': self.n_states**2,
            'emissions': self.n_states * self.n_features
        }

    @property
    def dof(self):
        return self.n_states ** 2 + self.n_states * self.n_features - self.n_states - 1

    def _update_B_params(self,X,log_gamma,theta):
        gamma = [torch.exp(gamma) for gamma in log_gamma]
        CategoricalEmissions.update_emission_params(self,X,gamma,theta)

    def check_sequence(self,sequence):
        return validate_sequence(sequence,True,self.n_features)

    def map_emission(self,x):
        return CategoricalEmissions.map_emission(self,x)

    def sample_B_params(self,X,seed=None):
        self._emission_matrix = CategoricalEmissions.sample_emissions_params(self,X,seed)





from setuptools import setup, find_packages #type: ignore

with open('README.md', encoding="utf-8") as handle:
    LONG_DESCRIPTION = handle.read()

setup(
    name='chadhmm',
    packages=find_packages(),
    version='0.3.4',
    description='Package for Hidden (Semi) Markov Models',
    author='GarroshIcecream',
    author_email='ad.pesek13@gmail.com',
    url='https://github.com/GarroshIcecream/ChadHMM',
    download_url = 'https://github.com/GarroshIcecream/ChadHMM/archive/refs/tags/v0.3.tar.gz',
    license='MIT',
    keywords=['Hidden Markov Models','Hidden Semi-Markov Models','hsmm','hmm','Gaussian Mixture Models', 'gmm'],
    install_requires=['torch','matplotlib','scikit-learn','numpy','prettytable'],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
    ]
)


from typing import Optional, Literal, List
from sklearn.cluster import KMeans # type: ignore
import torch
from torch.distributions import MultivariateNormal

from .base_emiss import BaseEmission # type: ignore
from ..utils import ContextualVariables, validate_means, validate_covars, fill_covars # type: ignore


class GaussianEmissions(BaseEmission):
    """
    Gaussian Distribution for HMM emissions.    
    
    Parameters:
    ----------
    n_dims (int):
        Number of mixtures in the model. This is equal to the number of hidden states in the HMM.
    n_components (int):
        Number of components in the mixture model.
    n_features (int):
        Number of features in the data.
    init_weights (bool):
        Whether to initialize the mixture weights prior to fitting.
    init_dist ():
        Distribution to use for initializing the mixture weights.
    alpha (float):
        Dirichlet concentration parameter for the prior over mixture weights.
    k_means (bool):
        Whether to initialize the mixture means using K-Means clustering.
    min_covar (float):
        Minimum covariance for the mixture components.
    covariance_type (COVAR_TYPES):
        Type of covariance matrix to use for the mixture components. One of 'spherical', 'tied', 'diag', 'full'.
    dims (Sequence[str]):
        Names of the mixtures in the model.
    seed (int):
        Random seed for reproducibility.
    device (torch.device):
        Device to use for computations.
    """

    COVAR_TYPES_HINT = Literal['spherical', 'tied', 'diag', 'full']

    def __init__(self, 
                 n_dims: int,
                 n_features: int,
                 params_init: bool = True,
                 k_means: bool = False,
                 covariance_type: COVAR_TYPES_HINT = 'full',
                 min_covar: float = 1e-3,
                 seed: Optional[int] = None,
                 device: Optional[torch.device] = None):
        
        super().__init__(n_dims, n_features, False, seed, device)

        self.min_covar = min_covar
        self.covariance_type = covariance_type
        self.k_means = k_means
        if params_init:
            self._means, self._covs = self.sample_emissions_params()
            
    def __str__(self):
        return f'GaussianEmissions(n_dims={self.n_dims}, n_features={self.n_features})'        

    @property
    def means(self) -> torch.Tensor:
        return self._means
    
    @means.setter
    def means(self, means: torch.Tensor):
        valid_means = validate_means(means, self.n_dims, self.n_features)
        self._means = valid_means.to(self.device)

    @property
    def covs(self) -> torch.Tensor:
        return self._covs

    @covs.setter
    def covs(self, new_covars: torch.Tensor):
        """Setter function for the covariance matrices."""
        valid_covars = validate_covars(new_covars, self.covariance_type, self.n_dims, self.n_features)
        self._covs = fill_covars(valid_covars, self.covariance_type, self.n_dims, self.n_features).to(self.device)

    @property
    def pdf(self) -> MultivariateNormal:
        return MultivariateNormal(self.means,self.covs)
    
    @property
    def params(self):
        return {'means': self.means, 
                'covs': self.covs}
    
    def map_emission(self,x):
        x_batched = x.unsqueeze(1).expand(-1,self.n_dims,-1)
        return self.pdf.log_prob(x_batched)
    
    def sample_emissions_params(self,X=None,seed=None):
        if X is not None:
            means = self._sample_kmeans(X, seed) if self.k_means else X.mean(dim=0).expand(self.n_dims,-1).clone()
            centered_data = X - X.mean(dim=0)
            covs = (torch.mm(centered_data.T, centered_data) / (X.shape[0] - 1)).expand(self.n_dims,-1,-1).clone()
        else:
            means = torch.zeros(size=(self.n_dims, self.n_features), 
                                dtype=torch.float64, 
                                device=self.device) 
            covs = self.min_covar + torch.eye(n=self.n_features, 
                                              dtype=torch.float64,
                                              device=self.device).expand((self.n_dims, self.n_features, self.n_features)).clone()

        return means, covs
    
    def update_emission_params(self,X,gamma,theta=None):
        self._means.copy_(self._compute_means(X,gamma,theta))
        self._covs.copy_(self._compute_covs(X,gamma,theta))
    
    def _sample_kmeans(self, X:torch.Tensor, seed:Optional[int]=None) -> torch.Tensor:
        """Sample cluster means from K Means algorithm"""
        k_means_alg = KMeans(n_clusters=self.n_dims, 
                             random_state=seed, 
                             n_init="auto").fit(X)
        return torch.from_numpy(k_means_alg.cluster_centers_).reshape(self.n_dims,self.n_features)

    def _compute_means(self,
                       X:List[torch.Tensor],
                       gamma:List[torch.Tensor],
                       theta:Optional[ContextualVariables]) -> torch.Tensor:
        """Compute the means for each hidden state"""
        new_mean = torch.zeros(size=(self.n_dims, self.n_features), 
                               dtype=torch.float64, 
                               device=self.device)
        
        denom = torch.zeros(size=(self.n_dims,1), 
                            dtype=torch.float64, 
                            device=self.device)
        
        for seq,gamma_val in zip(X,gamma):
            if theta is not None:
                # TODO: matmul shapes are inconsistent 
                raise NotImplementedError('Contextualized emissions not implemented for GaussianEmissions')
            else:
                new_mean += gamma_val.T @ seq
                denom += gamma_val.T.sum(dim=-1,keepdim=True)

        return new_mean / denom
    
    def _compute_covs(self, 
                      X:List[torch.Tensor],
                      gamma:List[torch.Tensor],
                      theta:Optional[ContextualVariables]) -> torch.Tensor:
        """Compute the covariances for each component."""
        new_covs = torch.zeros(size=(self.n_dims,self.n_features, self.n_features), 
                               dtype=torch.float64, 
                               device=self.device)
        
        denom = torch.zeros(size=(self.n_dims,1,1), 
                            dtype=torch.float64, 
                            device=self.device)

        for seq,gamma_val in zip(X,gamma):
            if theta is not None:
                # TODO: matmul shapes are inconsistent 
                raise NotImplementedError('Contextualized emissions not implemented for GaussianEmissions')
            else:
                gamma_expanded = gamma_val.T.unsqueeze(-1)
                diff = seq.expand(self.n_dims,-1,-1) - self.means.unsqueeze(1)
                new_covs += torch.transpose(gamma_expanded * diff,1,2) @ diff
                denom += torch.sum(gamma_expanded,dim=-2,keepdim=True)

        new_covs /= denom
        new_covs += self.min_covar * torch.eye(self.n_features, device=self.device)

        return new_covs


import torch
from abc import ABC, abstractmethod, abstractproperty
from typing import Optional, List, Dict
from torch.distributions import Distribution, MixtureSameFamily
from ..stochastic_matrix import WeightsMatrix
from ..utils import ContextualVariables, log_normalize


class MixtureEmissions(ABC):
    """
    Mixture model for HMM emissions. This class is an abstract base class for Gaussian, Poisson and other mixture models.
    """

    def __init__(self, 
                 n_dims: int,
                 n_features: int,
                 n_components: int,
                 alpha: float = 1.0,
                 init_weights: bool = True,
                 seed: Optional[int] = None,
                 device: Optional[torch.device] = None):
        
        self.n_dims = n_dims
        self.alpha = alpha
        self.n_components = n_components
        self.n_features = n_features
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') if device is None else device

        if init_weights:
            self._weights = self.sample_weights(seed)

    @property
    def weights(self) -> WeightsMatrix:
        try:
            return self._weights
        except AttributeError:
            raise AttributeError('Weights are not initialized')

    @weights.setter
    def weights(self, vector):
        assert (self.n_dims, self.n_components) == vector.shape, 'Matrix dimensions differ from HMM model'
        if isinstance(vector, WeightsMatrix):
            self._weights = vector
        elif isinstance(vector, torch.Tensor):
            self._weights = WeightsMatrix(n_states=self.n_dims, 
                                          n_components=self.n_components,
                                          matrix=vector,
                                          device=self.device)
        else:
            raise NotImplementedError(f'Expected torch Tensor or WeightsMatrix object, got {type(vector)}')
        
    @property
    def mixture_pdf(self) -> MixtureSameFamily:
        """Return the emission distribution for Gaussian Mixture Distribution."""
        return MixtureSameFamily(mixture_distribution = self.weights._dist,
                                 component_distribution = self.pdf)

    @abstractproperty
    def pdf(self) -> Distribution:
        """Return the emission distribution of Mixture."""
        pass

    @abstractproperty
    def params(self) -> Dict[str, torch.Tensor]:
        """Return the parameters of the Mixture."""
        pass

    @abstractmethod
    def sample_emissions_params(self, X:Optional[torch.Tensor]=None, seed:Optional[int]=None):
        """Sample the parameters of the Mixture."""
        pass

    @abstractmethod
    def update_emission_params(self, 
                               X:List[torch.Tensor], 
                               resp:List[torch.Tensor], 
                               theta:Optional[ContextualVariables]=None) -> None:
        """Update the parameters of the Mixture."""
        pass 
    
    def sample_weights(self, seed:Optional[int]) -> WeightsMatrix:
        """Sample the weights for the mixture."""
        return WeightsMatrix(n_states=self.n_dims,
                             n_components=self.n_components,
                             rand_seed=seed,
                             alpha=self.alpha,
                             device=self.device)    

    def map_emission(self, x:torch.Tensor) -> torch.Tensor:
        x_batched = x.unsqueeze(1).expand(-1,self.n_dims,-1)
        return self.mixture_pdf.log_prob(x_batched)
    
    def _compute_responsibilities(self, X:List[torch.Tensor]) -> List[torch.Tensor]:
        """Compute the responsibilities for each component."""
        resp_vec = []
        for seq in X:
            n_observations = seq.size(dim=0)
            log_responsibilities = torch.zeros(size=(self.n_dims,self.n_components,n_observations), 
                                               dtype=torch.float64, 
                                               device=self.device)

            for t in range(n_observations):
                log_responsibilities[:,:,t] = log_normalize(self.weights.matrix + self.pdf.log_prob(seq[t]),1)

            resp_vec.append(log_responsibilities)
        
        return resp_vec
    
    def _compute_weights(self, posterior:List[torch.Tensor]) -> torch.Tensor:
        log_weights = torch.zeros(size=(self.n_dims,self.n_components),
                                  dtype=torch.float64, 
                                  device=self.device)

        for p in posterior:
            log_weights += p.exp().sum(-1)
        
        return log_normalize(log_weights.log(),1)